#include "rayTracer.h"
#include "scene.h"
#include "camera.h"
#include "hit.h"
#include "ray.h"
#include "RTConfig.h"

#include <qfileinfo.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <QProgressDialog>
#include <QTime>

QTime globTime;


#define GL_BGRA 0x80E1
#define GL_TEXTURE_RECTANGLE_EXT 0x84F5

void RayTracer::renderImage() const
{
	int xres=scene()->camera().xResolution();
	int yres=scene()->camera().yResolution();

	globTime.start();

	QProgressDialog qpd("RenderImage", "Cancel", 0, xres*yres) ;
	qpd.setWindowModality(Qt::WindowModal);

	image_=QImage( xres, yres, QImage::Format_RGB32);


 
	int numrays=RTConfig::getInstance()->getNumPrimaryRays();
	float step=1.0f/(float)numrays;


	for (int j=0; j<yres; ++j)
		for (int i=0; i<xres; ++i){
	  
		  Color color(0,0,0);
		  for (float k=0; k<1; k+=step)
			  for (float l=0; l<1; l+=step) 
				color += this->rayColorWithRefraction(scene()->camera().getPixelRay(i+k,j+l), 0, 0, qglviewer::Vec(0,0,0));

		  color = color / (float)(numrays*numrays);
		  image_.setPixel(i, j, color.operator QRgb() );

		if(i==1){
			qpd.setValue(i+j*xres);
			if(qpd.wasCanceled())
				return;
		}
	  }



	////Copying images data into the texture/////
	if(!this->imgTexID)
		glGenTextures(1, &(this->imgTexID));

	RTConfig::getInstance()->previewTexID=this->imgTexID;

	glBindTexture(GL_TEXTURE_RECTANGLE_EXT, this->imgTexID);
	glTexParameteri(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glTexImage2D( GL_TEXTURE_RECTANGLE_EXT, 0, GL_RGBA8, xres, yres, 0, GL_BGRA, GL_UNSIGNED_BYTE, image_.bits() );
	/////////////

	int t=globTime.elapsed();
	if(t>0)
		std::cout<<"renderImage time: "<<t<<"\n";
}


// RayColor without refraction
Color RayTracer::rayColor(const Ray& ray, int numReb, qglviewer::Vec oinray) const
{
  Color res;
  Hit hit;

  if(numReb==0)
	  rayPath_.clear();

  if(scene()->intersect(ray, hit)){

	  // "Partie VII: Premire Image"
	  // Get only the diffuse color of the intersected object
	  // res = hit.material().diffuseColor();

	  // Test to check the normals
	  //res.r=abs(hit.normal().x);
	  //res.g=abs(hit.normal().y);
	  //res.b=abs(hit.normal().z);


	  // "Partie VIII: Eclairage"
	  // Compute the color of this pixel with all the lights of the scene
	  res = scene()->illuminatedColor(hit);

	  // "Partie X: Plusieurs rebonds"

	  //Segment creation for debugging display
	  rayPath_.push_back(Segment(ray, hit));

	  
	  if(numReb<maxReboundCount){

		  // Init the new ray which is the reflected ray
		  Ray newray;
		  newray.setStart(hit.intersection());
		  newray.setDirection( (ray.direction())+2*((-ray.direction())*hit.normal())*hit.normal() );

		  // Call the method rayColor itself with the new parameter
		  Color reboundcol = rayColor(newray, numReb+1, qglviewer::Vec(0,0,0));

		  // Add to the pixel's color the color of the new hit affected by the reflective color
		  res = res + reboundcol*hit.material().reflectiveColor();

	  }
  }

  // No rebound, display the background
  else if(numReb==0)
	res = scene()->backgroundColor();
  
  // End of the rebounds
  else{
		res=Color(0,0,0);
		rayPath_.push_back(Segment(ray));
  }

  // "Partie 11: Textures"
  // Test to check the coordinates of the texture
  //res.r=abs(hit.u());
  //res.g=abs(hit.v());
  //res.b=0;

  return res;
}



// RayColor with refraction
Color RayTracer::rayColorWithRefraction(const Ray& ray, int reflectDepth, int refractDepth, qglviewer::Vec oinray) const
{
  Color res;
  Hit hit;

  if(reflectDepth==0 && refractDepth==0)
	  rayPath_.clear();


  if(scene()->intersect(ray, hit)){

	  // "Partie VII: Premire Image"
	  // Get only the diffuse color of the intersected object
	  // res = hit.material().diffuseColor();

	  // Test to check the normals
	  //res.r=abs(hit.normal().x);
	  //res.g=abs(hit.normal().y);
	  //res.b=abs(hit.normal().z);


	

	  //Segment creation for debugging display
	  rayPath_.push_back(Segment(ray, hit));

	  
	  // "Extension: Refraction"
#if 0
	  // The ray was inside and now goes out
	  if(oinray.norm()>0){
		  if(RTConfig::getInstance()->testFlag){
			  Ray newray;
			  newray.setStart(hit.intersection());
			  newray.setDirection( oinray );

			  Color refractcol=rayColorWithRefraction(newray, reflectDepth, refractDepth); //refractDepth+1 ?

			  res = res + refractcol*hit.material().refractiveColor();
				/////
		  }else{
			  Ray newray;
			  newray.setStart(hit.intersection());
			  qglviewer::Vec i=ray.direction();
			  qglviewer::Vec n=-hit.normal();
			  float eta=1.0f/hit.material().refractiveIndex();


			  float cosi = (-i)*n;
			  float cost2 = 1.0 - eta * eta * (1.0 - cosi*cosi);
			  
			  if(cost2>0){
				  qglviewer::Vec refractRay = eta*i + ((eta*cosi - sqrt(abs(cost2))) * n);
				 // qglviewer::Vec refractRay= qglviewer::Vec(t.x* (cost2 > 0.0), t.y* (cost2 > 0.0), t.z* (cost2 > 0.0));
			
			  
				  newray.setDirection( refractRay );

				  Color refractcol = rayColorWithRefraction(newray, reflectDepth, refractDepth);
				  res = res+refractcol*hit.material().refractiveColor(); 
			  }
		  }

	  }else
#endif
	  {

	  // "Partie VIII: Eclairage"
	  // Compute the color of this pixel with all the lights of the scene
	  res = scene()->illuminatedColorWithRefraction(hit);

	  // "Partie X: Plusieurs rebonds"


#if 1
	  if(1 && hit.material().refractiveColor().intensity()>0){

///Fast fresnel

		float n1;	  		  
		float n2;
		qglviewer::Vec normal;

		if(oinray.norm()>0){
				n1=1.0f/hit.material().refractiveIndex();
				n2=1.0f;

				normal=-hit.normal().unit();
		}else{
				n1=1.0f;
				n2=1.0f/hit.material().refractiveIndex();
				normal=hit.normal().unit();
		}

		float eta = n1/n2; 

		float r0=((n1-n2)/(n1+n2))*((n1-n2)/(n1+n2));
		float fastFresnel= r0 + (1-r0)*pow(1.0f -(-ray.direction() * normal), 5);


		qglviewer::Vec reflection = (ray.direction())+2*((-ray.direction())*normal)*normal; //a moi

		qglviewer::Vec refraction;
			qglviewer::Vec i=ray.direction();
			qglviewer::Vec n=normal;
			//  float eta=hit.material().refractiveIndex();

			float cosi = (-i)*n;
			float cost2 = 1.0 - eta * eta * (1.0 - cosi*cosi);

			if(cost2>0){
				refraction = eta*i + ((eta*cosi - sqrt(abs(cost2))) * n);
			}else
				refraction=qglviewer::Vec(0,0,0);

		Color reflectCol(0,0,0);
		Color refractCol(0,0,0);

		Ray newray;
		newray.setStart(hit.intersection());

		if(reflectDepth < RTConfig::getInstance()->getMaxReflection() && oinray.norm()<=0 ){
			newray.setDirection(reflection);
			reflectCol = rayColorWithRefraction(newray, reflectDepth+1, refractDepth);

			res+=reflectCol*fastFresnel*hit.material().reflectiveColor();
		}

		if(refractDepth < RTConfig::getInstance()->getMaxRefraction() && refraction.norm()>0){
			newray.setDirection(refraction);
			if(oinray.norm()>0)
				refractCol = rayColorWithRefraction(newray, reflectDepth, refractDepth);
			else
				refractCol = rayColorWithRefraction(newray, reflectDepth, refractDepth+1, ray.direction());

			res+=refractCol*(1.0f-fastFresnel)*hit.material().refractiveColor();
		}
		
//Fresnel First version
#if 0
		  float index_external;	  		  
		  float index_internal;
		  qglviewer::Vec incom=ray.direction().unit();
		  qglviewer::Vec normal;


		  if(oinray.norm()>0){
				index_external=1.0f/hit.material().refractiveIndex();
				index_internal=1.0f;

				normal=-hit.normal().unit();
		  }else{
				index_external=1.0f;
				index_internal=1.0f/hit.material().refractiveIndex();
				normal=hit.normal().unit();
		  }

		  

		  float eta = index_external/index_internal; 
		  float cos_theta1 = incom * normal; 
		  float cos_theta2 = sqrt(1.0f - ((eta * eta) * ( 1.0f - (cos_theta1 * cos_theta1)))); 

		 // std::cout<<eta<<" "<<cos_theta1<<" "<<cos_theta2<<"\n";
		  //qglviewer::Vec reflection = (ray.direction())+2*((-ray.direction())*normal)*normal; //a moi
		  qglviewer::Vec reflection = incom - 2.0f * cos_theta1 * normal; //original
		  qglviewer::Vec refraction = (eta * incom) + (cos_theta2 - eta * cos_theta1) * normal; 

		  /*qglviewer::Vec refraction;
		  
				  qglviewer::Vec i=ray.direction();
				  qglviewer::Vec n=normal;
				//  float eta=hit.material().refractiveIndex();

				  float cosi = (-i)*n;
				  float cost2 = 1.0 - eta * eta * (1.0 - cosi*cosi);

				  if(cost2>0){
					   refraction = eta*i + ((eta*cosi - sqrt(abs(cost2))) * n);
				  }else
					  refraction=qglviewer::Vec(0,0,0);*/



		  float fresnel_rs = (index_external * cos_theta1 - index_internal * cos_theta2 ) / (index_external * cos_theta1 + index_internal * cos_theta2); 
		  float fresnel_rp = (index_internal * cos_theta1 - index_external * cos_theta2 ) / (index_internal * cos_theta1 + index_external * cos_theta2); 

		  float reflectance, transmittance;
		  if(oinray.norm()<=0){
			reflectance = (fresnel_rs * fresnel_rs + fresnel_rp * fresnel_rp) / 2.0f; 
			transmittance =((1.0f-fresnel_rs) * (1.0f-fresnel_rs) + (1.0f-fresnel_rp) * (1.0f-fresnel_rp)) / 2.0f;
		  }else{
			reflectance=0;
			transmittance=1.0f;
		  }

		 // std::cout<<eta<<" "<<cos_theta1<<" "<<cos_theta2<<" : "<<fresnel_rs<<" "<<fresnel_rp<<" ! "<<reflectance<<" "<<transmittance<<"\n";

		  Color reflectCol(0,0,0);
		  Color refractCol(0,0,0);

		  Ray newray;
		  newray.setStart(hit.intersection());

		  if(reflectDepth < RTConfig::getInstance()->getMaxReflection() && oinray.norm()<=0 ){
			newray.setDirection(reflection);
			reflectCol = rayColorWithRefraction(newray, reflectDepth+1, refractDepth);
			
			res+=reflectCol*reflectance*hit.material().reflectiveColor()*0.5f;
		  }

		  if(refractDepth < RTConfig::getInstance()->getMaxRefraction() && refraction.norm()>0){
			newray.setDirection(refraction);
			if(oinray.norm()>0)
				refractCol = rayColorWithRefraction(newray, reflectDepth, refractDepth);
			else
				refractCol = rayColorWithRefraction(newray, reflectDepth, refractDepth+1, ray.direction());

			res+=refractCol*transmittance*hit.material().refractiveColor()*0.5f;
		  }
#endif
//Fresnel V2
#if 0
		  bool bInside=oinray.norm()>0;
		  
				// material de type verre on va calculer le coefficient de reflexion de fresnel

				qglviewer::Vec vNormal;

				float fDensity1;
				float fDensity2;
				
				float fReflectance, fTransmittance;
				float fCosThetaI, fSinThetaI, fCosThetaT, fSinThetaT;
				

				if (bInside){
					// on considere que le rayon incident traverse un ether proche du vide (ou de l'air)
					// en thorie il faudrait d'abord calculer pour savoir si l'extrieur
					// se trouve dans un autre objet mais a dpasse le cadre de notre code.

					fDensity1 = 1.0f/hit.material().refractiveIndex();
					fDensity2 = 1.0f;

					vNormal=-hit.normal().unit();
				}
				else {
					fDensity1 = 1.0f;
					fDensity2 = 1.0f/hit.material().refractiveIndex();
					
					vNormal=hit.normal().unit();
				}

				float fViewProjection = ray.direction().unit() * vNormal;

				// ici on tient compte du fait que le dplacement de la lumiere
				// est symetrique, de l'observateur a la source de la lumiere ou de la source de lumiere
				// a l'observateur. On fait donc les calculs du coefficient en prenant en compte le rayon provenant
				// du point d'observation.
				fCosThetaI = fabsf(fViewProjection);

				if (fCosThetaI >= 0.999f) {
					// cas du rayon incident selon la normale a la surface
					fReflectance = (fDensity1 - fDensity2) / (fDensity1 + fDensity2);
					fReflectance = fReflectance * fReflectance;
					fSinThetaI = 0.0f;
					fSinThetaT = 0.0f;
					fCosThetaT = 1.0f;
				}else {
					fSinThetaI = sqrtf(1 - fCosThetaI * fCosThetaI);
					// le signe de Sin ThetaI n'a aucune importance, c'est le meme que celui de Sin ThetaT
					// et ils s'annihilent mutuellement dans le calcul des coefficients de Reflexion
					fSinThetaT = (fDensity1 / fDensity2) * fSinThetaI;
					if (fSinThetaT * fSinThetaT > 0.9999f)
					{
						// On a atteint l'angle critique au dela duquel
						// la surface est entierement reflexive
						fReflectance = 1.0f ;
						fCosThetaT = 0.0f;
					}
					else
					{
						fCosThetaT = sqrtf(1 - fSinThetaT * fSinThetaT);
						// tout d'abord la reflectance de la lumiere polaris dans le
						// plan orthogonal au plan de reflexion
						float fReflectanceOrtho = (fDensity2 * fCosThetaT - fDensity1 * fCosThetaI )
							/ (fDensity2 * fCosThetaT + fDensity1  * fCosThetaI);
						fReflectanceOrtho = fReflectanceOrtho * fReflectanceOrtho;
						// Ensuite la reflectance de la lumiere polaris dans le
						// plan parallele au plan de reflexion
						float fReflectanceParal = (fDensity1 * fCosThetaT - fDensity2 * fCosThetaI )
							/ (fDensity1 * fCosThetaT + fDensity2 * fCosThetaI);
						fReflectanceParal = fReflectanceParal * fReflectanceParal;

						// le coefficient de reflectance global est gal  la moyenne
						// des deux pour une lumire de base non polarise
						fReflectance =  0.5f * (fReflectanceOrtho + fReflectanceParal);
					}
				}

				fTransmittance = /*currentMat.refraction * */(1.0f - fReflectance);
				fReflectance = /*currentMat.reflection **/ fReflectance;
			
				////

				
			Color reflectCol(0,0,0);
			Color refractCol(0,0,0);

			qglviewer::Vec reflection = (ray.direction())+2*((-ray.direction())*vNormal)*vNormal;//incom - 2.0f * cos_theta1 * normal; 
		
			float eta = fDensity1/fDensity2; 
			qglviewer::Vec refraction;
				  qglviewer::Vec i=ray.direction();
				  qglviewer::Vec n=vNormal;
				//  float eta=hit.material().refractiveIndex();

				  float cosi = (-i)*n;
				  float cost2 = 1.0 - eta * eta * (1.0 - cosi*cosi);

				  if(cost2>0){
					   refraction = eta*i + ((eta*cosi - sqrt(abs(cost2))) * n);
				  }else
					  refraction=qglviewer::Vec(0,0,0);

			Ray newray;
			newray.setStart(hit.intersection());

			 if(reflectDepth < RTConfig::getInstance()->getMaxReflection() && oinray.norm()<=0 ){
				newray.setDirection(reflection);
				reflectCol = rayColorWithRefraction(newray, reflectDepth+1, refractDepth);
				
				res+=reflectCol*fReflectance*hit.material().reflectiveColor()*0.5f;
			  }

			  if(refractDepth < RTConfig::getInstance()->getMaxRefraction() && refraction.norm()>0){
				newray.setDirection(refraction);
				if(oinray.norm()>0)
					refractCol = rayColorWithRefraction(newray, reflectDepth, refractDepth);
				else
					refractCol = rayColorWithRefraction(newray, reflectDepth, refractDepth+1, ray.direction());

				res+=refractCol*fTransmittance*hit.material().refractiveColor()*0.5f;
			  }

#endif

	  }else{
			if(reflectDepth < RTConfig::getInstance()->getMaxReflection()) {
			  // Init the new ray which is the reflected ray
			  Ray newray;
			  newray.setStart(hit.intersection());
			  newray.setDirection( (ray.direction())+2*((-ray.direction())*hit.normal())*hit.normal() );

			  // Call the method rayColor itself with the new parameter
			  Color reboundcol = rayColorWithRefraction(newray, reflectDepth+1, refractDepth);

			  // Add to the pixel's color the color of the new hit affected by the reflective color
			  res = res + reboundcol*hit.material().reflectiveColor();
		  }
	  }
		  
#else
		  if(reflectDepth < RTConfig::getInstance()->getMaxReflection() &&  hit.material().reflectiveColor().intensity()>0 ) {
			  // Init the new ray which is the reflected ray
			  Ray newray;
			  newray.setStart(hit.intersection());
			  newray.setDirection( (ray.direction())+2*((-ray.direction())*hit.normal())*hit.normal() );

			  // Call the method rayColor itself with the new parameter
			  Color reboundcol = rayColorWithRefraction(newray, reflectDepth+1, refractDepth);

			  // Add to the pixel's color the color of the new hit affected by the reflective color
			  res = res + reboundcol*hit.material().reflectiveColor();
		  }

		 if(refractDepth < RTConfig::getInstance()->getMaxRefraction() &&  hit.material().refractiveColor().intensity()>0) {
			  // "Extension: Refraction"
			  Ray newray;
			  newray.setStart(hit.intersection());
			  
			  //Refraction computation
			  if(hit.material().refractiveColor().intensity()>0){
				  qglviewer::Vec i=ray.direction();
				  qglviewer::Vec n=hit.normal();
				  float eta=hit.material().refractiveIndex();


				  float cosi = (-i)*n;
				  float cost2 = 1.0 - eta * eta * (1.0 - cosi*cosi);

				  if(cost2>0){
					  qglviewer::Vec refractRay = eta*i + ((eta*cosi - sqrt(abs(cost2))) * n);
					  //qglviewer::Vec refractRay= qglviewer::Vec(t.x* (cost2 > 0.0), t.y* (cost2 > 0.0), t.z* (cost2 > 0.0));
					
					  newray.setDirection( refractRay );

					  Color refractcol = rayColorWithRefraction(newray, reflectDepth, refractDepth+1, ray.direction());
					  res = res+refractcol*hit.material().refractiveColor(); 
				  }
			  } 
		  }

#endif
	  }

  } // No rebound, display the background
  else{   // End of the rebounds
	  if( !(reflectDepth==0 && refractDepth==0))
		rayPath_.push_back(Segment(ray));

	  if(hit.directColor)
		  res=hit.color_;
	  /*else
		res = scene()->backgroundColor();*/

	  hit.directColor=false;
  }
 
 

  // "Partie 11: Textures"
  // Test to check the coordinates of the texture
  //res.r=abs(hit.u());
  //res.g=abs(hit.v());
  //res.b=0;

  return res;
}



void RayTracer::saveImage(const QString& name, bool overwrite) const
{
  QString fileName = name;
  QFileInfo info(fileName);

  // Append jpg extension if needed
#if QT_VERSION < 0x040000
  if (info.extension(false) != "jpg")
#else
  if (info.suffix() != "jpg")
#endif
    {
      fileName += ".jpg";
      info.setFile(fileName);
    }

  // Prevent overwriting an existing file
  if (info.exists() && !overwrite)
    if (QMessageBox::warning(NULL, "Overwrite file ?",
			     "File "+info.fileName()+" already exists.\nOverwrite ?",
			     QMessageBox::Yes,
			     QMessageBox::Cancel) == QMessageBox::Cancel)
      return;

  if (!image().save(fileName, "JPEG", 95))
    QMessageBox::information(NULL, "Error", "Error while saving "+fileName);
  else
    QMessageBox::information(NULL, "Saving done", "Image successfully saved in "+fileName);
}


// Draw the path of the ray between each rebound
void RayTracer::drawRayPath(){
	
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glLineWidth(5.0f);

	glBegin(GL_LINES);
	for(int i=0; i<this->rayPath_.size(); i++){
		glVertex3f(rayPath_[i].start.x, rayPath_[i].start.y, rayPath_[i].start.z);	
		glVertex3f(rayPath_[i].end.x, rayPath_[i].end.y, rayPath_[i].end.z);
	}
	glEnd();

	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
	glBegin(GL_LINES);
	for(int i=0; i<this->rayPath_.size(); i++){
		glVertex3f(rayPath_[i].end.x, rayPath_[i].end.y, rayPath_[i].end.z);	
		glVertex3f((rayPath_[i].end.x+rayPath_[i].endNormal.x*0.1), (rayPath_[i].end.y+rayPath_[i].endNormal.y*0.1), (rayPath_[i].end.z+rayPath_[i].endNormal.z*0.1));
	}
	glEnd();

	glPopAttrib();

}


void RayTracer::drawImage(float size){
	if(!this->imgTexID)
		return;

	if(RTConfig::getInstance()->getPreviewMode()!=RTConfig::RTC_PM_WINDOW)
		return;

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glDepthMask(1);


	// switch to projection mode
	glMatrixMode(GL_PROJECTION);
	// save previous matrix which contains the 
	//settings for the perspective projection
	glPushMatrix();
	// reset matrix
	glLoadIdentity();
	// set a 2D orthographic projection
	gluOrtho2D(0, 10, 0, 10);
	//glOrtho(0, 10, 0, 10, 0.005f, sqrt(3.0));

	// invert the y axis, down is positive
	glScalef(1, -1, 1);
	// mover the origin from the bottom left corner
	// to the upper left corner
	glTranslatef(0, -10, 0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	//this->resTex->bind();
	glBindTexture(GL_TEXTURE_RECTANGLE_EXT, this->imgTexID);
	glEnable(GL_TEXTURE_RECTANGLE_EXT);


	glDisable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); //Back to front

	glColor4f(1,1,1, 0.5);

	glScalef(size,size,1);

	//glTranslatef(0,0.2,0);
	glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3f(0,0,0.1);

		glTexCoord2f(scene()->camera().xResolution(),0);
		glVertex3f(1,0,0.1);

		glTexCoord2f(scene()->camera().xResolution(), scene()->camera().yResolution());
		glVertex3f(1,1,0.1);

		glTexCoord2f(0,scene()->camera().yResolution());
		glVertex3f(0,1,0.1);


	glEnd();

	glBindTexture(GL_TEXTURE_RECTANGLE_EXT, 0);

	glDisable(GL_TEXTURE_RECTANGLE_EXT);
	//ModelView
	glPopMatrix();
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glPopAttrib();

}