#include "scene.h"
#include "sphere.h"
#include "plane.h"
#include "triangle.h"
#include "triangleMesh.h"
#include "cubeMap.h"
#include <qfile.h>

using namespace qglviewer;

Scene::~Scene(){
// Clear the list of objects & lights
  for(int i=0; i<objectList.count(); i++)
		delete objectList[i];
  objectList.clear();

  for(int i=0; i<lightList.count(); i++)
	  delete lightList[i];
  lightList.clear();
}


void Scene::draw() const
{

	// Draw the "real" lights in the scene (ambient, diffuse and specular)
	glEnable(GL_LIGHTING);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);


	//glEnable(GL_DEPTH_TEST);

	if(RTConfig::getInstance()->drawLight && RTConfig::getInstance()->drawHelpers){	
		for(int i=0; i<lightList.count(); i++){
		  Light *l=((Scene *)this)->lightList.at(i);
		  l->draw();
		}
	}

	// draw the all the objects of the scene
    for(int i=0; i<objectList.count(); i++){
      Object *o=((Scene *)this)->objectList.at(i);
      o->draw();
    }
	

	// Put the camera according to the radius of the scene
	//camera_.draw(0.1);


}

void Scene::addObject(Object* o)
{
  objectList.append(o);
}


void Scene::addLight(Light* l)
{
  lightList.append(l);
}


#ifndef IO_ReadOnly
#define IO_ReadOnly QIODevice::ReadOnly
#endif

void Scene::loadFromFile(const QString& filename)
{
  QDomDocument doc( "scene" );

  QFile file( filename );
  if ( !file.open( IO_ReadOnly ) )
    return;

  if ( !doc.setContent( &file ) ) {
    file.close();
    return;
  }

  file.close();
  
  // Clear the list of objects & lights
  for(int i=0; i<objectList.count(); i++)
		delete objectList[i];
  objectList.clear();

  for(int i=0; i<lightList.count(); i++)
	  delete lightList[i];
  lightList.clear();

  // print out the element names of all elements that are direct children
  // of the outermost element.
  QDomElement docElem = doc.documentElement();

  
  //GPURT special modes
  this->specialMode=(int)docElem.attribute("specialMode", "0").toInt();

  QDomNode n = docElem.firstChild();
  while( !n.isNull() ) {
    QDomElement e = n.toElement(); // try to convert the node to an element.
    if( !e.isNull() ) {
		Object *otmp;
	  if(e.tagName()=="Sphere"){
		otmp=new Sphere();
		otmp->initFromDOMElement(e);
		this->addObject(otmp);
	  }else if(e.tagName()=="Plane"){
		otmp=new Plane();
		otmp->initFromDOMElement(e);
		this->addObject(otmp);
	  }else if(e.tagName()=="Triangle"){
		otmp=new Triangle();
		otmp->initFromDOMElement(e);
		this->addObject(otmp);
	  }else if(e.tagName()=="TriangleMesh"){
		otmp=new TriangleMesh();
		otmp->initFromDOMElement(e);
		this->addObject(otmp);
	  }if(e.tagName()=="CubeMap"){
		otmp=new CubeMap();
		otmp->initFromDOMElement(e);
		this->addObject(otmp);
	  }

	  else if(e.tagName()=="Camera"){
		  camera_.initFromDOMElement(e);
	  }
	  else if(e.tagName()=="BackgroundColor"){
		  this->backgroundColor_.initFromDOMElement(e);
	  }
	  else if(e.tagName()=="AmbientLight"){
		Light *ltmp=new AmbientLight();
		ltmp->initFromDOMElement(e);
		this->addLight(ltmp);
	  }
	  else if(e.tagName()=="DirectionalLight"){
		Light *ltmp=new DirectionalLight();
		ltmp->initFromDOMElement(e);
		this->addLight(ltmp);
	  }
	  else if(e.tagName()=="PointLight"){
		Light *ltmp=new PointLight();
		ltmp->initFromDOMElement(e);
		this->addLight(ltmp);
	  }

	  else if(e.tagName()=="SphereArray"){

		  int sx=e.attribute("sizeX", "0").toInt();
		  int sy=e.attribute("sizeY", "0").toInt();
		  int sz=e.attribute("sizeZ", "0").toInt();

		  Color cools[]={Color(1.0f, 0.0f, 0.0f), Color(0.0f, 1.0f, 0.0f), 
				Color(0.0f, 0.0f, 1.0f), Color(1.0f, 0.8f, 0.0f)};


		  for(int k=0; k<sz; k++)
			for(int j=0; j<sy; j++)
			  for(int i=0; i<sx; i++){
				Sphere *stmp=new Sphere();
				stmp->initFromDOMElement(e);
				((Material&)stmp->material()).setDiffuseColor(cools[rand()%4]);
				((Frame&)stmp->frame()).setPosition(Vec(i*1.0f/(sx-1), j*1.0f/(sy-1), k*1.0f/(sz-1))-Vec(0.5f, 0.5f, 0.5f));

				this->addObject(stmp);
			  }
	  }
    }
    n = n.nextSibling();
  }
}


// Compute the extrema of the scene, ie gives the corners of the bounding box
void Scene::compMinMax(){
  if(objectList.count()<=0)
    return;
  
  Object *otmp=objectList.at(0);

  mini.x=otmp->frame().position().x-otmp->boundingRadius();
  mini.y=otmp->frame().position().y-otmp->boundingRadius();
  mini.z=otmp->frame().position().z-otmp->boundingRadius();

  maxi.x=otmp->frame().position().x+otmp->boundingRadius();
  maxi.y=otmp->frame().position().y+otmp->boundingRadius();
  maxi.z=otmp->frame().position().z+otmp->boundingRadius();
  
  for(int i=1; i<objectList.count() ; i++){

    Vec tmposmin=objectList.at(i)->frame().position();
    tmposmin.x-=objectList.at(i)->boundingRadius();
    tmposmin.y-=objectList.at(i)->boundingRadius();
    tmposmin.z-=objectList.at(i)->boundingRadius();

    if(tmposmin.x<mini.x)
      mini.x=tmposmin.x;
    if(tmposmin.y<mini.y)
      mini.y=tmposmin.y;
    if(tmposmin.z<mini.z)
      mini.z=tmposmin.z;

    Vec tmposmax=objectList.at(i)->frame().position();  
    tmposmax.x+=objectList.at(i)->boundingRadius();
    tmposmax.y+=objectList.at(i)->boundingRadius();
    tmposmax.z+=objectList.at(i)->boundingRadius();

    if(tmposmax.x>maxi.x)
      maxi.x=tmposmax.x;
    if(tmposmax.y>maxi.y)
      maxi.y=tmposmax.y;
    if(tmposmax.z>maxi.z)
      maxi.z=tmposmax.z;
  }
}


// Compute the center of the scene
Vec Scene::center() const
{
  //TODO: Flag pour pas recalculer
  ((Scene *)this)->compMinMax();
  return (mini+maxi)/2.0f;
}


// Compute the radius of the scene
float Scene::radius(const Vec& center) const
{
  ((Scene *)this)->compMinMax();
  return (maxi-mini).norm()/2.0f;
}


// Test the intersection of a ray with all the objects of the scene
bool Scene::intersect(const Ray& ray, Hit& hit) const {
  bool res=false;
  for(int i=0; i<objectList.count(); i++){
      Object *o=((Scene *)this)->objectList.at(i);
	  if( o->intersect(ray, hit) )
		  res=true;
  }
  return res;
}

// Compute all the lights for one intersected pixel
Color Scene::illuminatedColor(const Hit& hit) const{
  
  Color res(0,0,0);

  // Compute the direction of the "eye" (for the specular light)
  qglviewer::Vec eye = (camera().frame().position() - hit.intersection());
  eye.normalize();

  for(int i=0; i<lightList.count(); i++){
      Light *l=((Scene *)this)->lightList.at(i);
	  
	  // "Partie IX: Les ombres"
	  Hit rayhit;
	  if(l->visibleFrom(hit.intersection(), this, rayhit))
		res = res + l->illuminatedColor( eye, hit);
  }

  return res;
}


// Compute all the lights for one intersected pixel with Refraction
Color Scene::illuminatedColorWithRefraction(const Hit& hit) const{
  
  Color res(0,0,0);

  // Compute the direction of the "eye" (for the specular light)
  qglviewer::Vec eye = (camera().frame().position() - hit.intersection());
  eye.normalize();

  for(int i=0; i<lightList.count(); i++){
      Light *l=((Scene *)this)->lightList.at(i);
	  
	  // "Partie IX: Les ombres"
	  Hit rayhit;
	  if(l->visibleFrom(hit.intersection(), this, rayhit))
		res = res + l->illuminatedColor( eye, hit);

	  // "Extension: Refraction"
	  else {
		  if(rayhit.material().refractiveColor().intensity()>0) {
			  res = res+(l->illuminatedColor( eye, hit)*rayhit.material().refractiveColor() * rayhit.material().diffuseColor(rayhit.u(), rayhit.v()) );//+ 
              //(1.0f-rayhit.material().refractiveColor())*rayhit.material().diffuseColor(rayhit.u(), rayhit.v()) ) ;
		  }
	  }
  }

  return res;
}